/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

Authors:    Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
			Telecomunication Networks Group
			Politecnico di Torino
			Torino, Italy
			mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file
    Defines template algorithms used for generation of various neighbour
	and chunk selection methods. 
 */
#ifndef _algorithms_h_
#define _algorithms_h_

#include <algorithm>
#include "random.h"
#include "globals.h"

/// Helper templates for specifying filter predicates
template <class T>
struct Predicate1 {
	bool (*func)(const T);
	inline Predicate1(bool (*func)(const T)) : func(func) {};
	inline bool accept(T entity) const {
		return func == NULL || (*func)(entity);
	}
};

template <class T, class ARGT>
struct Predicate2 {
	bool (*func)(const T, ARGT);
	ARGT arg;
	inline Predicate2(bool (*func)(const T, ARGT), ARGT arg) : func(func), arg(arg) {};
	inline bool accept(T entity) const {
		return func == NULL || (*func)(entity, arg);
	}
};


/// Choose a random entity within the given iterator range from among those that aren't in the
/// "exclusion" set (if provided) and which satisfy
/// the given filter predicate (which has a bool method 'accept' taking a T as a parameter).
/// The supplied function "weight" is the acceptance probability relative to other entities,
/// so that for example a entity with twice the weight of some other entity is
/// twice as likely to be selected.
/// If filter is NULL, all entities are evaluated. If weight is NULL, all entities
/// have value 1 (so a random choice is made from all those passing the filter).
/// Returns not_found_val if no entities satisfy the predicate.
/// Three versions are provided -- the first takes a refce to a predicate class for
/// filtering the entities, the second takes a pointer to a filter function and the
/// third takes a filter function with an additional argument.
template < class T, class IT, class PRED, class SETTYPE >
inline T choose_weighted_entity(
				IT first, ///< Iterator that references the first candidate entity
				IT end,   ///< Iterator that references the last candidate entity
				SETTYPE* exclusions, ///< Pointer to set of entities to exclude. NULL for empty set.
				const PRED& pred,   ///< Predicate function to apply to candidates
				double (*weight)(const T), ///< Weight (value) function to apply to an entity
				T not_found_val)  ///< Value to return if no candidate found
{
	std::vector<double> cdf;	// Cumulative Density Function
	std::vector<T> entities;
	T best = not_found_val;
	double total_weight = 0;
	for (IT it = first; it != end; ++it) {
		T entity = *it;
		bool excluded = exclusions != NULL && 
			std::find(exclusions->begin(), exclusions->end(), entity) != exclusions->end();
		if (!excluded && pred.accept(entity)) {
			double wt = weight == NULL ? 1 : (*weight)(entity);
			total_weight += wt;
			cdf.push_back(total_weight);
			entities.push_back(entity);
		}
	}
	if (!cdf.empty()) {
		if (total_weight == 0) {  // Special case when all weights are zero
			size_t index = (size_t) uniform(0, cdf.size(), &gSeed);
			best = entities[index];
		}
		else {
			double ran = uniform(0, total_weight, &gSeed);  // The "coin toss" selection value
			best = entities[std::lower_bound(cdf.begin(), cdf.end(), ran) - cdf.begin()];
		}
	}
	return best;
}


template <class T, class IT, class SETTYPE >
inline T choose_weighted_entity(
				IT first,
				IT end,
				SETTYPE* exclusions,
				bool (*accept)(const T),
				double (*weight)(const T),
				T not_found_val)
{
	return choose_weighted_entity(first, end, exclusions, Predicate1<T>(accept), weight, not_found_val);
}

template <class T, class IT, class ARGT, class SETTYPE >
inline T choose_weighted_entity(
				IT first,
				IT end,
				SETTYPE* exclusions,
				bool (*accept)(const T, ARGT),
				ARGT arg, 
				double (*weight)(const T),
				T not_found_val)
{
	return choose_weighted_entity(first, end, exclusions, Predicate2<T, ARGT>(accept, arg), weight, not_found_val);
}


/// Choose, within the given iterator range (where 'end' is the "(last+1)th",
/// from among all entities satisfying the given filter filter predicate
/// (which has a bool method 'accept' taking a T as a parameter).
/// that entity with the greatest value, as determined by the supplied 'value' function.
/// If more than one entity has the same maximum value, a random one is selected.
/// If filter is NULL, all entities are evaluated. If value is NULL, all entities
/// have value 1, so a random choice is made from all those passing the filter.
/// Returns not_found_val if no entities satisfy the predicate.
/// Two versions are provided -- the second allows the filter function to take an
/// additional parameter.
template < class T, class IT, class PRED, class SETTYPE > 
inline T choose_best_entity(
				IT first,
				IT end,
				SETTYPE* exclusions,
				const PRED& pred,
				double (*value)(const T),
				T not_found_val)
{
	std::vector<T> best_set;
	T best = not_found_val;
	double max_value = -std::numeric_limits<double>::infinity();
	for (IT it = first; it != end; ++it) {
		T entity = *it;
		bool excluded = exclusions != NULL && 
			std::find(exclusions->begin(), exclusions->end(), entity) != exclusions->end();
		if (!excluded && pred.accept(entity)) {
			double val = value == NULL ? 1 : (*value)(entity);
			if (val > max_value) {
				max_value = val;
				best_set.clear();
				best_set.push_back(entity);
			}
			else if (val == max_value) {
				best_set.push_back(entity);
			}
		}
	}
	if (!best_set.empty()) {
		best = best_set[(size_t) (uniform(0, best_set.size(), &gSeed))];
	}
	return best;
}

	
template <class T, class IT, class SETTYPE > 
inline T choose_best_entity(
			IT first,
			IT end,
			SETTYPE* exclusions,
			bool (*accept)(const T),
			double (*value)(const T),
			T not_found_val)
{
	return choose_best_entity(first, end, exclusions, Predicate1<T>(accept), value, not_found_val);
}

template <class T, class IT, class ARGT, class SETTYPE > 
inline T choose_best_entity(
			IT first,
			IT end,
			SETTYPE* exclusions,
			bool (*accept)(const T, ARGT),
			ARGT arg,
			double (*value)(const T),
			T not_found_val)
{
	return choose_best_entity(first, end,  exclusions, Predicate2<T, ARGT>(accept, arg), value, not_found_val);
}




#endif // _algorithms_h_
