/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Maurizio Munafo', Marco Mellia, Gloria Bogliolo, Ana Paula Couto,
		  Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file
	Implementation of the AOPeer -- "Adaptive Overlay Peer" -- a peer that
	continually adapts its local neighbourhood in order to maintain a small
	excess of unused outlinks in a given time window.
 */

#include <sstream>
#include <cmath>
#include <algorithm>
#include "aopeer.h"
#include "misc.h"
#include "chunk.h"
#include "link.h"
#include "globals.h"
#include "random.h"
#include "overlay.h"
#include "myassert.h"
#include "videostream.h"

static size_t MIN_NEIGHBOURHOOD_INC = 1;   // Minimum no. of neighbours to get when growing
static double INITIAL_GROWTH_FACTOR = 0.4; // Initial neighbourhood growth fraction
static double FINAL_GROWTH_FACTOR = 0.1;   // Final neighbourhood growth fraction (see next line)
static size_t CULL_CYCLES_TILL_FINAL = 15; // Number of "time windows" before final growth rate reached

// Definitions of the various static class variables
// See aopeer.h for explanations.

size_t AOPeer::scan_period = 0;
size_t AOPeer::low_tide_percent = 10;
size_t AOPeer::high_tide_percent = 30;
size_t AOPeer::min_in_links = 1;
size_t AOPeer::min_out_links = 1;
double AOPeer::min_connect_time = 100;
double AOPeer::max_connect_time = 1000;
ReplenishModeType AOPeer::replenish_mode = BAD_VALUE;
AOPeer::SchedType AOPeer::cs_desirability = AOPeer::SQRT_OUTDEG;
AOPeer::SchedType AOPeer::n_desirability = AOPeer::SQRT_OUTDEG;

// Convert from a string representation of the desired scheduling type (or what is really
// the desirability function) to its SchedType enumeration equivalent.
AOPeer::SchedType getSchedType(const std::string type) {
	static std::string typeStrings[] = {"OUTDEG", "SQRT_OUTDEG", "UPLOAD_BW", "SQRT_UPLOAD_BW"};
	static AOPeer::SchedType types[] = {AOPeer::OUTDEG, AOPeer::SQRT_OUTDEG, 
		AOPeer::UPLOAD_BW, AOPeer::SQRT_UPLOAD_BW};
	static const int NUM_TYPES = sizeof(typeStrings) / sizeof(std::string);
	std::string* p = std::find(typeStrings, &typeStrings[NUM_TYPES], type);
	if (p == &typeStrings[NUM_TYPES]) {
		fatal(std::string("Unknown scheduling type: ") + type);
	}
	return types[p - typeStrings];
}

// Initialise static fields
void AOPeer::init_statics() {
	if (config.has_key("ReplenishMode")) {
		if (config.getString("ReplenishMode") == "BA_REPLENISH") {
			replenish_mode = BA_REPLENISH;
		}
		else if (config.getString("ReplenishMode") == "RANDOM_REPLENISH") {
			replenish_mode = RANDOM_REPLENISH;
		}
	}
	scan_period = config.has_key("ScanPeriod") ? config.getInt("ScanPeriod") : 0;
	low_tide_percent = config.has_key("LowTidePercent") ? config.getInt("LowTidePercent") : 10;
	high_tide_percent = config.has_key("HighTidePercent") ? config.getInt("HighTidePercent") : 30;
	min_in_links = config.has_key("MinInLinks") ? config.getInt("MinInLinks") : 1;
	min_out_links = config.has_key("MinOutLinks") ? config.getInt("MinOutLinks") : 1;
	min_connect_time = config.has_key("MinConnectTime") ? config.getDouble("MinConnectTime") : 100;
	max_connect_time = config.has_key("MaxConnectTime") ? config.getDouble("MaxConnectTime") : 1000;
	n_desirability = config.has_key("NeighbourDesirability") ? 
		getSchedType(config.getString("NeighbourDesirability")) : SQRT_OUTDEG;
	cs_desirability = config.has_key("ChunkSchedDesirability") ?
		getSchedType(config.getString("ChunkSchedDesirability")) : SQRT_OUTDEG;
}


// Bandwidth squared (a possible "desirability" function)
double sq_bandwidth(PLink plink) {
	return plink->to()->upload_bandwidth() * plink->to()->upload_bandwidth();
}

// Sqrt(bandwidth) (a possible "desirability" function)
double sqrt_bandwidth(PLink plink) {
	return sqrt(plink->to()->upload_bandwidth());
}

// Return the desirability of a candidate link for chunk scheduling
double chunk_sched_desirability(PLink plink) {
	PPeer p = plink->to();
	switch (AOPeer::cs_desirability) {
		case AOPeer::SQRT_OUTDEG:
			return std::sqrt((double) p->num_links_out());
		case AOPeer::OUTDEG:
			return p->num_links_out();
		case AOPeer::UPLOAD_BW:
			return p->upload_bandwidth();
		case AOPeer::SQRT_UPLOAD_BW:
			return std::sqrt(p->upload_bandwidth());
	}
	return std::sqrt((double) p->num_links_out());
}

/// Estimate of how desirable this peer is as a neighbour. Usually
/// this is a measure of its upload bandwidth, but could be something else.
double AOPeer::desirability() {
	switch (n_desirability) {
		case SQRT_OUTDEG:
			return std::sqrt((double) num_links_out());
		case OUTDEG:
			return num_links_out();
		case UPLOAD_BW:
			return upload_bandwidth();
		case SQRT_UPLOAD_BW:
			return std::sqrt(upload_bandwidth());
	}
	return std::sqrt((double) num_links_out());
}

// Non-member desirability function. Again a measure used in neighbourhood
// selection.
double desirability(PPeer p) {
	return ((AOPeer*) p)->desirability();
}

/// Chunk scheduler -- same as LATEST_BA but using a variety of possible desirability functions.
void AOPeer::schedule_new_chunk()
{
	bool done = false;
	ChunkIterator it = window()->begin();	// Iterator over chunks, latest first
	while (it != window()->end() && !done  && upload_bandwidth() > 0 
			&& num_links_out() > 0 && !p_upload_busy) {
		PChunk chunk = *it;
		PLink neighbour = choose_weighted_neighbour(NULL, &needs_chunk_and_not_busy, chunk, &chunk_sched_desirability);
		if (neighbour != NULL) {
			chunks_out_since_last_scan++;
			schedule_chunk(chunk, neighbour);
			done = true;
		}
		else {
			++it;
		}
	}
}

/// Overridden version of receive_chunk in order to schedule the neighbourhood
/// scanner at regular intervals.
void AOPeer::receive_chunk(PChunk chunk, PPeer from) {
	chunks_since_last_scan++;
	if (p_is_churner && p_tot_chunks_recd == p_churn_time) {
		Peer::receive_chunk(chunk, from);
		reset_all_links();
		chunks_since_last_scan = 0;
		p_churn_time = p_tot_chunks_recd + int(uniform(min_connect_time, max_connect_time, &gSeed));
	}
	else {
		if (scan_period != 0 &&  chunks_since_last_scan == scan_period) {
			adjust_neighbourhood();
			chunks_since_last_scan = 0;
		}
		Peer::receive_chunk(chunk, from);
	}
}

/// Delete all links in and out of this Peer and reconnect to the overlay
void AOPeer::reset_all_links()
{
	printf("Peer %d resetting at t = %.3lf \n", id(), gCurrentTime);
	while(first_out_link() != last_out_link()) {
		delete_link(first_out_link());
	}

	while (p_feeders.begin() != p_feeders.end()) {
		(*p_feeders.begin())->delete_link(this);
	}

	p_po->connect_peer(this);
}


/// Called at the end of every time window to adjust the neighbourhood.
void AOPeer::adjust_neighbourhood() {
	// Scan through all links and make a list of ones not used since last scan
	std::vector<PLink> unused_links;
	for (LinkIterator it = first_out_link(); it != last_out_link(); ++it) {
		if (num_scan_cycles != 0 && (*it)->stats()->throughput() == 0) {
			unused_links.push_back(*it);
		}
		(*it)->stats()->clear();
	}

	if (num_scan_cycles > 0) {  // The first scan cycle isn't over a full time period, so ignore

		// If the number of unused links is less than the specified low-tide level, top up.

		size_t num_used = num_links_out() - unused_links.size();
		size_t low_water = std::max(1, (int) (num_used * low_tide_percent / 100.0 + 0.5));
		if (unused_links.size() < low_water) {
			double growth_factor = INITIAL_GROWTH_FACTOR + (FINAL_GROWTH_FACTOR - INITIAL_GROWTH_FACTOR) *
				 std::min(1.0, num_scan_cycles/(double) CULL_CYCLES_TILL_FINAL);
			size_t num_wanted = std::max((size_t) 1, (size_t) (num_links_out() * growth_factor));
			grow_neighbourhood(num_wanted);
		}

		// If the number of unused links is greater than the specified high-tide level, cull
		// some of the unused links to get the unused number to mid-tide level, but with the
		// constraint that we don't cull a link to a peer that is already at its minimum
		// in-link level. This may mean that we don't achieve the desired cull target or
		// conceivably don't cull anything. Undesirable, but what to do about it ... ?

		size_t high_water = (int) (num_used * high_tide_percent / 100.0 + 0.5);
		if (unused_links.size() > high_water) {
			shrink_neighbourhood(num_used, unused_links);
		}
	}

	num_scan_cycles++;
	chunks_out_since_last_scan = 0;
}

// Shrink the neighbourhood by culling a random subset of the provided
// unused links, ignoring links to neighbours already at their minimum in-degree.
// Target neighbourhood size is the mean of high and low tide levels.
void AOPeer::shrink_neighbourhood(size_t num_used, std::vector<PLink>& unused_links) {
	LinkList candidates;
	PeerList candidate_peers;  // Another representation of candidates
	size_t ideal_num_unused = std::max(1,
		(int) (num_used * (high_tide_percent + low_tide_percent) / 200.0 + 0.5));
	size_t num_to_cull = std::min(unused_links.size() - ideal_num_unused,
		num_links_out() - min_out_links);

	// Check which of the unused links are actually cullable, i.e., are
	// links to the source or to a peer with more than its minimum in-degree.

	for (size_t i = 0; i < unused_links.size(); i++) {
		PPeer target = unused_links[i]->to();
		if (target->id() == 0 || target->num_links_in() > min_in_links) {
			candidates.push_back(unused_links[i]);
			candidate_peers.push_back(unused_links[i]->to());
		}
	}

	// Stochastically select a desirable subset to keep, and delete the rest.
	// We use the same stochastic selection function as we use when growing the
	// neighbourhood.

	num_to_cull = std::min(candidates.size(), num_to_cull);
	size_t num_to_keep = candidates.size() - num_to_cull;
	PeerList keepers;
	stochastic_select(num_to_keep, candidate_peers, keepers);
	for (size_t i = 0; i < candidates.size(); ++i) {
		if (std::find(keepers.begin(), keepers.end(), candidates[i]->to()) == keepers.end()) {
			delete_link(candidates[i]);
		}
	}
}

/// Grow the neighbourhood by selecting a suitable subset of all the neighbours' neighbours.
void AOPeer::grow_neighbourhood(size_t num_wanted) {
	PeerSet candidates;
	for (LinkIterator it = first_out_link(); it != last_out_link(); ++it) {
		PeerSet neighbours_neighbours = (*it)->to()->neighbour_set();
		candidates.insert(neighbours_neighbours.begin(), neighbours_neighbours.end());
	}
	PeerSet my_neighbours = neighbour_set();
	my_neighbours.insert(this);   // Include outselves in the neighbourhood to avoid links to self
	my_neighbours.insert(p_po->peer(0));  // Include source to avoid adding links back to it
	PeerList result(candidates.size());
	PeerIterator last_candidate = std::set_difference(
		candidates.begin(), candidates.end(),
		my_neighbours.begin(), my_neighbours.end(),
		result.begin()
	);
	size_t num_available = last_candidate - result.begin();
	result.resize(num_available);
	size_t num_to_take = std::min(num_wanted, num_available);
	PeerList new_neighbours;
	stochastic_select(num_to_take, result, new_neighbours);
	for (size_t i = 0; i < new_neighbours.size(); i++) {
		add_link(new_neighbours[i]);
	}
}

// Predicate for sorting (size_t, double) pairs in decreasing order of the double value
bool decreasingWeight(std::pair<size_t, double> a, std::pair<size_t, double> b) {
	return a.second > b.second;
}

// Select into "selection" a desirable subset of "num_to_take" neighbours
// from the given "candidates" list according
// to the current replenish mode (either bandwidth aware or random).
void AOPeer::stochastic_select(size_t num_to_take, PeerList& candidates, PeerList& selection) {

	// First compute individual and total desirability of candidate peers

	size_t num_available = candidates.size();
	std::vector< std::pair<size_t, double> > indices; // Indices and associated candidate weight

	double total_value = 0;
	for (size_t i = 0; i < candidates.size(); i++) {
		double weight = ((AOPeer*) candidates[i])->desirability();
		total_value += weight;
		double randomised_wt = uniform(0, weight, &gSeed);
		indices.push_back(std::pair<size_t, double>(i, randomised_wt));
	}

	if (replenish_mode == BAD_VALUE) {
		fatal("AOPeer: Replenish mode not specified");
	}
	else if (replenish_mode == RANDOM_REPLENISH || total_value == 0) {
		std::random_shuffle(indices.begin(), indices.end());
	}
	else { // Desirability-weighted selection
		// Sort indices_and_weights in decreasing order of randomised weight
		// std::sort(indices.begin(), indices.end(), decreasingWeight);	
		size_t n = std::min(num_to_take, num_available);
		for (size_t i = 0; i < n; i++) { 
			PPeer sel = choose_weighted_entity(candidates.begin(), candidates.end(),
				(std::set<PPeer>*) NULL, (bool (*)(PPeer)) NULL, &::desirability, (PPeer) NULL);
			myassert(sel != NULL, "AOPeer::stochastic_select: logic error");
			selection.push_back(sel);
			candidates.erase(std::find(candidates.begin(), candidates.end(), sel));
		}
		return;
	}

	// Now just copy items from the candidates list to the result list in the order
	// indicated by the permuted index list.

	size_t n = std::min(num_to_take, num_available);
	for (size_t i = 0; i < n; i++) {
		selection.push_back(candidates[indices[i].first]);
	}
}

/// Enable or disable churning for this peer.
void AOPeer::set_churn(bool churner) {
	p_is_churner = churner;
	p_churn_time = p_is_churner ? int(uniform(min_connect_time, max_connect_time, &gSeed)) : 0;
}
