/*********************************************************************/
/* Copyright (C) 2008  NEC Research Laboratories

Authors:    Ettore Ferranti
            Ettore.Ferranti@nw.neclab.eu
	    NEC Research Laboratories
	    Heidelberg, Germany

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

#include <vector>

#include "asnetwork.h"
#include "router.h"
#include "as.h"
#include "path.h"

/// default constructor, initialise the AS ids to 0
ASNetwork::ASNetwork()
{
    nextId = 0;
}

/// Compute the path latency *within the  network only* of the given path.
/// This is defined as the latency between the two ASes to which the routers are attached.
Latency ASNetwork::network_latency(Path path)
{
    PRouter rFrom = path.from->host()->router();
    PRouter rTo = path.to->host()->router();
    return this->getHops(rFrom, rTo);
}

/// Assign the specified router to the specified AS
void ASNetwork::assignRouter(PRouter router, PAS as)
{
    if (asMap.count(as->getID()) <= 0)
    {
        fatal("Attempt made to get non-existent as");
    }
    this->add_router(router);
    routerMap[router->getId()] = as->getID();
    as->addRouter(router);
}

/// returns the number of AS hops between two routers.
Latency ASNetwork::getHops(PRouter router1, PRouter router2)
{
    size_t id1 = router1->getId();
    size_t id2 = router2->getId();
    if (routerMap.count(id1) <= 0 || routerMap.count(id2) <= 0)
    {
        fatal("Attempt made to get a non-existent router");
    }
    AS_ID as1 = routerMap[id1];
    AS_ID as2 = routerMap[id2];
    if (as1 == as2) return 0;
    else return asMap[as1]->getLatency(as2);
}

/// given a peer A and a list of other peers, it orders the list according to
/// the latency between A and each one of the other peers, minimum latency first.
std::vector<PEER_LATENCY*> ASNetwork::orderList(PPeer source, std::vector<PPeer> list)
{
    std::vector <PEER_LATENCY*> lats;
    for (std::vector<PPeer>::iterator it = list.begin(); it!=list.end(); ++it)
    {
        PPeer other = *it;
        PEER_LATENCY* tmp = new PEER_LATENCY(other,getHops(source->host()->router(),other->host()->router()));
        lats.push_back(tmp);
    }
    std::sort(lats.begin(),lats.end(),pointer_less_than<PEER_LATENCY>());
    return lats;
}

/// create an AS and add it to the current network.
PAS ASNetwork::add_as()
{
    AS_ID nowid = nextId;
    nextId++;
    PAS tmp = new AS(nowid);
    asMap[nowid] = tmp;
    return tmp;
}

/// returns the pointer to the AS with the specified id.
PAS ASNetwork::as(AS_ID as)
{
    if (asMap.count(as) <= 0)
    {
        fatal("Attempt made to get non-existent as");
    }
    return asMap[as];
}

/// create a (bidirectional) link between the two specified ASes with the specified latency.
void ASNetwork::linkAs(AS_ID id1, AS_ID id2, Latency latency)
{
    if (asMap.count(id1) <= 0 || asMap.count(id1) <= 0 || id1 == id2)
    {
        fatal("Attempt made to get non-existent as");
    }
    asMap[id1]->addLink(id2, latency);
    asMap[id2]->addLink(id1, latency);
}



