/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file
	Implements the MBWOverlay class, which is an abstract overlay of
	multibandwidth peers. This class defines the set of peers, while
	its subclasses define how those peers are interconnected.
 */

#include <cstdlib>
#include <cmath>
#include "types.h"
#include "myassert.h"
#include "peer.h"
#include "mbwoverlay.h"
#include "random.h"
#include "misc.h"
#include "globals.h"
#include "propertymap.h"

// Define the MBWOverlay: a overlay with a set of peers chosen from 4 different bandwidth sets
// ========================================================================================

MBWOverlay::MBWOverlay() : Overlay() { }
// Method to define the peers of an MBW overlay.
// Defines a source peer with a bandwidth given by the config SourceUploadBW parameter if given
// or as a typical class 1 peer if not. Then defines a total of num_peers normal peers
// with classes, frequencies and bandwidths as specified by the PeerBandwidths config value.

void MBWOverlay::set_peers(size_t num_peers) {
	const double TOLERANCE = 0.000001;
	std::vector< std::vector<double> > bandwidths = config.getDoubleArray2D("PeerBandwidths");
	size_t num_classes = bandwidths.size();
	double bw_max_delta = config.getDouble("BandwidthFractVariation");

	// First generate the source peer (peer 0, of class 0).

	double src_upload_bandwidth = 0;
	if (config.has_key("SourceUploadBW")) {
		src_upload_bandwidth = config.getDouble("SourceUploadBW");
	}
	else {
//		src_upload_bandwidth = bandwidths[0][1] * (1 + uniform(-bw_max_delta, bw_max_delta, &gSeed));
		src_upload_bandwidth = bandwidths[0][1];
	}
	PPeer new_peer = Peer::make_src(src_upload_bandwidth, 0, 0);
	PeerID id = add_peer(new_peer); 

	// Now allocate all the "normal" peers

	size_t num_allocated = 0;
	double fract_total = 0;
	for (size_t iclass = 0; iclass < num_classes; iclass++) {
		std::vector<double> curr_class = bandwidths[iclass];
		if (curr_class.size() != 3) {
			fatal("MBWOverlay: a bandwidth specification for a peer class didn't have exactly 3 elements");
		}
		double fract = curr_class[0];
		double mean_upload = curr_class[1];
		double mean_download = curr_class[2];
		if (fract < 0 || fract > 1.0) {
			fatal("MBWOverlay: a bandwidth specification had an invalid fraction-of-occurrence value");
		}
		fract_total += fract;
		size_t num_in_class = (size_t) (num_peers * fract);

		// Treat the last class specially to make sure the right total number of peers is created
		if (iclass == num_classes - 1) { 
			if (fabs(fract_total - 1) > TOLERANCE) {
				fatal("MBWOverlay: the bandwidth class fractions did not add to 1");
			}
			num_in_class = num_peers - num_allocated; 
		}
		num_allocated += num_in_class;

		for (size_t i = 0; i < num_in_class; i++) {
			double upload_bandwidth = mean_upload * (1 + uniform(-bw_max_delta, bw_max_delta, &gSeed));
			double download_bandwidth = mean_download * (1 + uniform(-bw_max_delta, bw_max_delta, &gSeed));
			new_peer = Peer::make_peer(upload_bandwidth, download_bandwidth, iclass+1);
			id = add_peer(new_peer); 
		}
	}
}

