/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

Authors:    Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
			Telecomunication Networks Group
			Politecnico di Torino
			Torino, Italy
			mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file
	Declares the base Network class. In principle, a Network manages
	the transfer of information between two HOSTs (q.v). However, in the
	current implementation, events associated with chunk transmission are
	all managed by the HOSTs and the only role of the network class is
	to provide to the hosts the network latency, i.e., the additional
	delay, above and beyond the bandwidth induced delays between peers and
	router and the additional host-router and router-host latencies.
 */

#ifndef _NETWORK_H_
#define _NETWORK_H_

#include <vector>
#include <string>
#include "types.h"

/// The different types of networks.
/// The values in this enum must be matched, one for one, with the names in the
/// extern gNetworkTypeNames array.
enum NetworkType {
	AS_AWARE_NET,		///< AS aware network
	BAD_NET		///< Sentinel value (illegal) -- marks end of valid values
};

extern std::string gNetworkTypeNames[];  ///< String representations of network types

//static PNetwork p_current_network = NULL;

/// The Network class, which currently is concerned only with providing
/// to the HOSTs latency information on a particular path between two hosts
/// (and so might, for example, model various forms of congestion).
class Network
{
	std::vector<PRouter> n_routers;
	std::vector<PHost> n_hosts;
        static NetworkType n_network_type;  ///< Defines what subclass of Network is generated by the network factory.
        
public:
	Network() {};				///< Default constructor
	RouterID add_router(PRouter);
	HostID register_host(PHost);
	PRouter router(RouterID router);
	PHost host(HostID host);
    static PNetwork n_current_network;

	/// Compute the path latency *within the  network only* of the given path.
	/// This is defined as the latency between the two routers only.
	/// In the base class this is zero. Subclasses may override.
	virtual Latency network_latency(Path path);

    /// Define the type of network to be generated by a call to the network factory method,
	/// create_network.
	static void set_network_type(NetworkType type) { n_network_type = type; }

	/// Decide whether a chunk being scheduled for delivery should actually
	/// be lost in the network.
	bool lose_chunk(PChunk chunk, Path path);

    /// Set the type of peer to be generated by the peer factory.
    //static void set_network_type(NetworkType network_type) { p_network_type = network_type; }
    
	// Network FACTORY METHOD
	// ===================
	static PNetwork make_network();
};

#endif // _NETWORK_H_
