/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Maurizio Munafo', Marco Mellia, Richard Lobb, Gloria Bogliolo, 
/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

Authors:    Maurizio Munafo', Marco Mellia, Gloria Bogliolo, 
            Ana Paula Couto, Richard Lobb
			Telecomunication Networks Group
			Politecnico di Torino
			Torino, Italy
			mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file peer.h
	Defines the structure of a Peer in a overlay, plus operations
	on it. Note however that Peer is an abstract class; the specific chunk-
	scheduling algorithm, defined by method schedule_new_chunk.
 */

#ifndef _PEER_H_
#define _PEER_H_

#include "types.h"
#include "window.h"
#include "stats.h"
#include "host.h"
#include "path.h"
#include "link.h"
#include "chunk.h"
#include "misc.h"
#include <limits>
#include <vector>
#include "algorithms.h"
#include "overlay.h"

/// The different types of peers, which currently correspond one for one
/// with chunk-scheduling algorithms.
/// For each of these, there needs to be a separate subclass of Peer with its
/// own specific schedule_new_chunk method. See peertypes.h/c
/// The values in this enum must be matched, one for one, with the names in the
/// extern gPeerTypeNames array.
enum PeerType {
	LATEST_RANDOM,		///< Lastest chunk, random peer
	LATEST_BA,			///< Latest chunk, bandwidth-aware peer selection (semi-random)
	CONFIGURABLE_PEER,	///< A re-configurable peer with a wide range of chunk-schedulers
	AO_PEER,			///< The adaptive-overlay peer: continually adjusts its neighbourhood
	ANA,				///< An experimental peer for use by Ana Paula
	TESTING_PEER,		///< A peer used by anyone in testing
	AS_AW_PEER,			///< An AS-network aware peer
        ROX,                            ///< An IDR-prioritized schedule peer for use by Rossella 
	BAD					///< Sentinel value (illegal) -- marks end of valid values
};

extern std::string gPeerTypeNames[];  ///< String representations of chunk-sched algorithms

/// The Peer class, representing a peer (peer) in an overlay network.

class Peer
{

private:
	PeerID p_id;					///< Peer ID (set by overlay when peer is added to it)
	PHost p_phost;					///< Pointer to associated host
	LinkList p_out_links;			///< List of outgoing links
	size_t p_peer_class;			///< Which class of peer this is
	Window p_window;				///< List of all current chunks that can be sent by a peer
	PeerStats p_stats;				///< Node statistics (app dependent so a separate class)
	static PeerType p_peer_type;	///< The type of peer to be generated by make_peer

protected:
	POverlay p_po;					///< Pointer to the overlay to which this peer belongs
	PeerSet p_feeders;				///< The set of all peers that have links to this peer
	bool p_upload_busy;				///< true if a peer is sending some chunk
	size_t p_tot_chunks_recd;		///< Total number of chunks received by this peer
	bool p_is_churner;				///< True iff this peer randomly disconnects and reconnects
	size_t p_churn_time;			///< Time, in chunks received, at which peer next "churns".
	
public:
	/// Set the type of peer to be generated by the peer factory.
	static void set_peer_type(PeerType peer_type) { p_peer_type = peer_type; }

	/// Default constructor for a Peer -- just initialises all fields to their default
	/// values. Must be called via a superclass constructor,
	/// as Peer is an abstract class.
	Peer()
		: p_po(NULL), p_phost(NULL), p_tot_chunks_recd(0),
		  p_upload_busy(false), p_peer_class(0), p_is_churner(false),
		  p_churn_time(0)
	{ }

	// Constructor for a Peer, for use when the underlay network is not being used.
	// This version constructs both the peer and an associated host. The host is not attached
	// to any network and serves only as a respository for bandwidth information. When sent
	// a chunk, it routes it to the target via the other host.
	// Must be called via a superclass constructor, as Peer is an abstract class.
	Peer(Bandwidth upload_bandwidth, Bandwidth download_bandwidth, size_t peer_class = 0,
		 Latency upload_latency = 0, Latency download_latency = 0)
		: p_po(NULL), p_peer_class(peer_class), p_tot_chunks_recd(0),
		  p_upload_busy(false), p_is_churner(false),
		  p_churn_time(0) 
	{
		p_phost = Host::create_host(upload_bandwidth, download_bandwidth, upload_latency, download_latency);
	}

	// GETTER methods
	// ==============

	/// The ID of this peer
	inline PeerID id()
	{ return p_id; }

	/// Is this peer the source?
	inline bool is_source()
	{ return id() == p_po->source(); }

	/// Pointer to the host associated with this peer
	inline PHost host()
	{ return p_phost; }

	/// Get the upload bandwidth of this peer, defined as the bits-per-second rate
	/// from the peer's host to the network (i.e., the router).
	inline Bandwidth upload_bandwidth()
	{ return p_phost->upload_bandwidth(); }

	/// Get the download bandwidth of this peer, defined as the bits-per-second rate
	/// from the network to the peer's host to the network (i.e., the router).
	inline Bandwidth download_bandwidth()
	{ return p_phost->download_bandwidth(); }

	/// The number of links out from this peer. Is essentially a count
	/// of the number of calls to add_link.
	inline size_t num_links_out()
	{ return p_out_links.size(); }

	/// The class of this peer. This is usually a bandwidth class but it's use
	/// is over to the subclasses.
	inline size_t get_class()
	{ return p_peer_class; }

	/// The number of links into this peer.
	inline size_t num_links_in()
	{ return p_feeders.size(); }

	/// True if this peer has neither received nor is in the process of
	/// receiving the specified chunk.
	inline bool needs_chunk(PChunk pchunk){ 
		return window()->needs_chunk(pchunk);
	}

	/// Latest chunk
	inline ChunkID latest(){
		PChunk l = window()->latest();
		return l ? l->id() : 0; 
	}

	/// Latest chunk, including pending ones
	inline ChunkID latest_w_pending(){
		PChunk l = window()->latest_w_pending();
		return l ? l->id() : -1; 
	}

	/// True if this peer's download capacity is currently saturated.
	inline bool download_busy() { 
		return host()->download_BW_avail() == 0;
	}

	/// Return a pointer to the statistics record for this peer.
	inline PPeerStats stats()
	{ return &p_stats; }

	/// Return whether or not this peer is a churner
	inline bool get_churner()
	{ return p_is_churner; }
	
	/// SETTER METHODS
	/// =============

	/// Set the peer id to the specified value
	inline void set_id(PeerID id) { p_id = id; }

	/// Set the overlay to the specified value
	inline void set_overlay(POverlay poverlay) { p_po = poverlay; }

	/// Set the host to the specified value
	inline void set_host(PHost host) { p_phost = host; }

	/// Called by another peer to notify this peer that a chunk
	/// has been scheduled for delivery to it.
	inline void notify_pending(PChunk chunk) {
		window()->set_pending(chunk);
	}

	/// Link access
	/// ===========
	/// Access to links should usually be via the for_each_link, which works
	/// for normal meshes and for fully-connected meshes. The iterators work
	/// only with normal meshes.

	/// Apply a function object to each 

	/// An iterator referencing the first out link of this peer.
	inline LinkIterator first_out_link()
	{ return p_out_links.begin(); }

	/// An iterator referencing the "last + 1" out link of this peer.
	inline LinkIterator last_out_link()
	{ return p_out_links.end(); }

	/// Add a link to a new neighbour.
	void add_link(PPeer dest);

	/// Delete the specified link from the neighbourhood
	virtual LinkIterator delete_link(LinkIterator it);

	/// Remove the given link
	void delete_link(PLink link);

	/// Remove the link to the given peer
	void delete_link(PPeer dest);

	/// Add a new peer to the set of known feeders
	void add_feeder(PPeer source);

	/// Remove a peer from the set of known feeders
	void remove_feeder(PPeer source);

	/// A set of all the neighbours of this peer, i.e., all peers to which it has outlinks
	PeerSet neighbour_set();

	/// Reset the peer back to its simulator-start state
	void clear();

	/// Reset the peer statistics gathering but not the link stats (which are needed, e.g.
	/// for adaptive overlay configuration).
	void reset_peer_stats();

	// Chunk SCHEDULING
	// ================

	// The main chunk sending method of a peer.
	// According to the current chunk-scheduling strategy
	// selects a chunk from the peer's window and sends it to one of the peer's
	// neighbours, if possible. If not possible, returns quietly without having
	// done anything. This method must be implemented by the subclasses.
	virtual void schedule_new_chunk() = 0;

	/// Called by the peer to send a chunk to another peer.
	/// Transmission will be via the peer's associated host, if there is one,
	/// of direct to the other peer if not.
	void send_chunk(PChunk chunk, Path path);

	/// Notify the peer that the chunk it was sending to the given peer has now
	/// been delivered or, at least, that the send is complete.
	/// [It is possible that the chunk was in error and discarded
	/// by the destination or that it has been held up elsewhere
	/// in the network.]
	/// This clears the peer's "up busy" status then calls the
	/// peer's try_to_send method to initiate sending
	/// of a new chunk if possible.
	void chunk_sent(PChunk chunk, PPeer dest);

	/// Send a new chunk (or create a ready to send event) if the peer
	/// is not busy.
	void try_to_send();

	/// Notify the peer that chunk sending has been finished and chunks that arrived 
        /// contemporarily have already been arrived.
        /// This is a workaround that makes results obtained by the event-based simulator
	/// equivalent to results of a cycle based symulator.
	void ready_to_send();

	/// Notify the peer that a new chunk has arrived for it.
	/// This clears the peer's "download busy" status, discards
	/// the chunk with a certain error probability, then calls the
	/// peer's schedule_new_chunk method, which is required to be implemented
	/// by the various subclasses of Peer.
	virtual void receive_chunk(PChunk chunk, PPeer from);

	/// Handle the chunk timeout event. This event is generated at network level
	/// whenever a chunk can't be delivered because the destination router->host
	/// link is saturated. 
	void chunk_timeout(PChunk chunk);


	/// A pointer to the peer's sliding window of chunks available for
	/// delivery to other peers.
	inline PWindow window()
	{ return &p_window; }

protected:
	// Support methods for chunk scheduling.
	// -------------------------------------

	/// Choose a random neighbour from among those that are not in the exclusion vector
	/// and which satisfy the given filter function 'accept'.
	/// The supplied function "weight" is the acceptance probability relative to other neighbours,
	/// so that for example a neighbour with twice the weight of some other neighbours is
	/// twice as likely to be selected.
	/// If exclusions is NULL, that parameter is ignored.
	/// If filter is NULL, all neighbours are evaluated. If weight is NULL, all neighbours
	/// have value 1 (so a random choice is made from all those passing the filter).
	/// Returns NULL if no neighbours satisfy the accept function.
	/// Two versions of the function are provided; the second allows the accept function to
	/// take a second parameter, e.g. the chunk_id that the peer should need.
	inline PLink choose_weighted_neighbour(std::vector<PLink>* exclusions,
			bool (*accept)(const PLink), double (*weight)(const PLink)) {
		return choose_weighted_entity(first_out_link(), 
									  last_out_link(),
									  exclusions, 
									  accept,
									  weight,
									  (PLink) NULL);
	}

	template <class ARGT>
	inline PLink choose_weighted_neighbour(std::vector<PLink>* exclusions, 
			bool (*accept)(const PLink, ARGT), ARGT arg, double (*weight)(const PLink)) {
		return choose_weighted_entity(  first_out_link(),
										last_out_link(),
										exclusions,
										accept,
										arg,
										weight,
										(PLink) NULL);
	}

	/// Choose, from among all neighbours not in the given "exclusions" set and that
	/// satisfying the given filter function 'accept', that neighbour 
	/// with the greatest value (as determined by the supplied 'value' function).
	/// If more than one neighbour has the same maximum value, a random one is selected.
	/// If filter is NULL, all neighbours are evaluated. If value is NULL, all neighbours
	/// have value 1 (so a random choice is made from all those passing the filter).
	/// Returns NULL if no neighbours satisfy the accept function.
	/// Two versions of the function are provided; the second allows the accept function to
	/// take a second parameter.
	inline PLink choose_best_neighbour(std::vector<PLink>* exclusions,
				bool (*accept)(const PLink), double (*value)(const PLink)) {
			return choose_best_entity(	first_out_link(),
										last_out_link(),
										exclusions,
										accept,
										value,
										(PLink) NULL);
	}

	template <class ARGT>
	inline PLink choose_best_neighbour(std::vector<PLink>* exclusions,
			bool (*accept)(const PLink, ARGT), ARGT arg, double (*value)(const PLink)) {
		return choose_best_entity(	first_out_link(),
									last_out_link(),
									exclusions,
									accept,
									arg,
									value,
									(PLink) NULL);
	}

	// The next group of methods are the various chunk-selection strategies. See the
	// corresponding neighbour-selection functions for explanations of the meanings
	// of all parameters. 
	// All return a value of NULL if no chunk is found.

	/// Choose a chunk using a weighting function. Similar to choose_weighted_neighbour (q.v.)
	inline PChunk choose_weighted_chunk(std::vector<PChunk>* exclusions,
			bool (*accept)(PChunk), double (*weight)(PChunk)) {
		return choose_weighted_entity(	window()->begin(),
										window()->end(),
										exclusions,
										accept,
										weight,
										(PChunk) NULL);
	}

	template <class ARGT>
	inline PChunk choose_weighted_chunk(std::vector<PChunk>* exclusions,
			bool (*accept)(const PChunk, ARGT), ARGT arg, double (*weight)(const PChunk)) {
		return choose_weighted_entity(	window()->begin(),
										window()->end(),
										exclusions,
										accept,
										arg,
										weight,
										(PChunk) NULL);
	}

	/// Choose the best chunk. Similar to choose_best_neighbour (q.v.)
	inline PChunk choose_best_chunk(std::vector<PChunk>* exclusions,
				bool (*accept)(const PChunk), double (*value)(const PChunk)) {
			return choose_best_entity(	window()->begin(),
										window()->end(), 
										exclusions,
										accept,
										value, 
										(PChunk) NULL);
	}

	template <class ARGT>
	inline PChunk choose_best_chunk(std::vector<PChunk>* exclusions,
			bool (*accept)(const PChunk, ARGT), ARGT arg, double (*value)(const PChunk)) {
		return choose_best_entity(	window()->begin(),
									window()->end(),
									exclusions,
									accept,
									arg,
									value,
									(PChunk) NULL);
	}

	// Schedule the given chunk for transmission on the given link
	void schedule_chunk(PChunk pchunk, PLink current_link);


public:

	// Peer FACTORY METHOD
	// ===================
	static PPeer make_peer(
	    Bandwidth upload_bandwidth,		///< The download_bandwidth of the peer
		Bandwidth download_bandwidth,	///< The upload_bandwidth of the peer
		size_t peer_class = 0			///< The class of the peer
	);

	static PPeer make_src(
	    Bandwidth upload_bandwidth,		///< The download_bandwidth of the peer
		Bandwidth download_bandwidth,	///< The upload_bandwidth of the peer
		size_t peer_class = 0			///< The class of the peer
	);

};

// Selector and weighting functions for peer and chunk selection
// =============================================================
// For use with choose_best_neighbour and choose_weighted_neighbour.

/// Selector for a neighbour that doesn't have the given chunk and isn't busy
bool needs_chunk_and_not_busy(const PLink link, PChunk chunk);

/// Returns the upload bandwidth of the link destination.
double peer_uploadBW(const PLink link);

/// Returns the download bandwidth of the link destination.
double peer_downloadBW(const PLink link);

/// Returns the age of the latest chunk of the link destination.
double peer_latestChunk(const PLink link);

// Weighting function for tit-for-tat like stategies.
double peer_down_up_ratio(const PLink link);

// Weighting function for BAELp algorithm.
double peer_baelp(const PLink link);

/// Weighting function that rewards the most recent chunks.
double chunk_recentness(PChunk chunk);

/// Weighting function that rewards the most recent chunks and priority.
double chunk_recentness_priority(PChunk chunk);

/// Weighting function that rewards chunk priority.
double chunk_priority(PChunk chunk);

/// Weighting function that rewards chunk with the lowest deadline value
double chunk_deadline(PChunk chunk);

#endif // _PEER_H_
