#include <cstdlib>
#include <cctype>
#include <algorithm>
#include "propertymap.h"
#include "misc.h"
#include <sstream>

// Build a set of allowable property map keys, each in lower case.
// If you wish to add your own keys to the property map, you must also
// add them to the "keys" array in the following function. Case is
// irrelevant -- they are converted to lower case by the function.
static std::set<std::string> make_key_set() {
	char* keys[] = {
		"ChunkSize","InterFrameTime","VideoRate","PlayoutDelay","MaxTreeDegree","ProbChunkError",
		"ChunkTimeoutInterval","TimeJitter","ContemporaryDeliveries","ChunkDelayCounters",
		"LinksFile","LinksAtEndFile","DegreeFile","DegreeAtEndFile",
		"OutOfSequenceFile","ThroughputFile","TotalChunkLossesPerPeerFile",
		"UploadRatePeerFile","UploadTotalPeerFile","DegreeOverTimeFile",
		"DelayHistogramsFile","DelayCdfFile","ChunkArrivalsFile","ChunkDelaysFile",
		"FilenameSuffix","FilenamePrefix",
		"InputDegreeFile","InputLinkFile","ConfigPeerConfig","ConfigPeerSourceConfig","DLIncrement","NetworkType",
		"HostType","OverlayType","PeerType","VideoStreamType","OverlaySeed","NumPeers",
		"Degree","NumChunks","PeerBandwidths","BandwidthFractVariation",
		"SourceUploadBW",
		"ScanPeriod","LowTidePercent","ConfigDumpFile",
		"ReplenishMode","StartStats","StopStats","StatusReportInterval","HighTidePercent",
		"MinInLinks", "MinOutLinks",
		"ChurnPercent", "MinConnectTime", "MaxConnectTime",
		"ChunkSchedDesirability", "NeighbourDesirability",
		"FECn","FECk", "WeightBAELp",
		"PriorityClassCount",
		"DLIncrDiff"
	};
	std::set<std::string> result;
	for (size_t i = 0; i < sizeof(keys)/sizeof(char*); i++) {
		std::string s = keys[i];
		std::transform(s.begin(), s.end(), s.begin(), (int(*)(int))std::tolower);
		result.insert(s);
	}
	return result;
}

// List of known (allowed) keys.
std::set<std::string> PropertyMap::known_keys = make_key_set();

// Utility method to issue a fatal error if an error occurs during parsing.
// The two parameter strings are simply concatenated.
void bad_input(std::string message, std::string extraInfo) {
	std::string error("PropertyMap error. ");
	error += message + extraInfo;
	fatal(error.c_str());
}

// Utility function to return the string s with all white space removed
// from both ends. Additionally, if the trimmed string begins with '"',
// then that character plus the corresponding terminating '"', is stripped.
std::string trim(std::string s) {
	while (s.size() > 0 && s[0] == ' ') {
		s.erase(s.begin());
	}
	while (s.size() > 0 && isspace(s[s.size()-1])) {
		s.erase(--s.end());
	}
	if (s.size() > 0 && s[0] == '"') {
		if (s[s.size()-1] != '"') {
			bad_input("PropertyMap: unterminated string value ", s);
		}
		s = s.substr(1,s.size()-2);
	}
	return s;
}

// Return a new string with all white space, opening and closing '"'s
// and comments stripped.
std::string stripCommentsAndSpaces(std::string s) {
	size_t commentStart = s.find("//");
	if (commentStart != std::string::npos) {
		s = s.substr(0, commentStart);
	}
	return trim(s);
}

// Extract a token representing a double value from the given string,
// returning the token. 
std::string nextDoubleToken(std::string& s) {
	s = trim(s);
	std::string token;
	if (s.size() >= 3 && s.substr(0,3) == "INF") {
		token = "INF";
		s = s.substr(3);
	}
	else {
		if (s.size() >= 1 && (s[0] == '+' || s[0] == '-')) {
			token = s[0];
			s = s.substr(1);
		}
		while (s.size() > 0 && (isdigit(s[0]) || s[0] == '.')) {
			token += s[0];
			s = s.substr(1);
		}
	}
	return token;
}

/// Load property map from a file. The input file is a text file
/// consisting of lines of the form 'name', 'name = value' or
/// 'name = "value"' where name is a sequence of alphanumeric or
/// underscore characters and value is any character sequence,
/// except that in the first version both leading and trailing white
/// space will be stripped.  Spaces around the '=' are optional, as are
/// leading spaces. Trailing comments, beginning with "//", are permitted
/// and are stripped before the (name, value) pair is inserted into the map.
void PropertyMap::load(const char* config_filename) {
	std::ifstream file(config_filename);
	std::string line, rawline;
	if (file.fail()) {
		fatal("Missing configuration file 'config.txt'. Copy 'config.example.txt' to 'config.txt' and personalize it!");
	}
	while (!std::getline(file, rawline).fail()) {
		process_line(rawline);
	}
}

/// Dump the contents of the map to a given file.
void PropertyMap::dump(FILE* file) {
	for (std::map<std::string, std::string>::iterator it = map.begin(); it != map.end(); ++it) {
		fprintf(file, "%s = \"%s\"\n", it->first.c_str(), it->second.c_str());
	}
	fclose(file);
}

/// Process an input line from the config file or an arg value from the command line
void PropertyMap::process_line(std::string rawline, bool allow_overwrite) {
	std::string line = stripCommentsAndSpaces(rawline);
	if (line.size() == 0) return;
	size_t end_of_name = line.find_first_not_of("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_");
	std::string name, value;
	if (end_of_name == std::string::npos) {  // Check for the special case of name by itself
		name = line;
	}
	else {	// Normal case. Get value.
		name = line.substr(0, end_of_name);
		std::string rest = trim(line.substr(end_of_name));
		if (rest.size() == 0 || rest[0] != '=') {
			bad_input("PropertyMap:: Error in parsing input line ", line);
		}
		value = trim(rest.substr(1));
	}

	if (name.size() == 0) {
		bad_input("Invalid property name in line or arg ", line);
	}
	std::transform(name.begin(), name.end(), name.begin(), (int(*)(int))std::tolower); // Horrible!
	if (!allow_overwrite && map.find(name) != map.end()) {
		bad_input("PropertyMap: Multiple definition of symbol ", name);
	}
	if (known_keys.find(name) == known_keys.end()) {
		bad_input("PropertyMap: Definition of unknown property ", name);
	}
	map[name] = value;
}


/// Process the command line args as property values. Aborts with fatal error if anything
/// goes wrong.
void PropertyMap::process_command_line_args(size_t argc, char* argv[]) {
	for (size_t i = 1; i < argc; i++) {
		process_line(argv[i], true /* allows replacing values in file with values from command */);
	}
}

/// True iff given symbol exists in the map
bool PropertyMap::has_key(std::string name) const {
	std::transform(name.begin(), name.end(), name.begin(), (int(*)(int))std::tolower); // Horrible!
	return map.find(name) != map.end();
}

/// Return the value of the given string in the property map.
/// Aborts (fatal error) if given string (i.e., key) isn't present in the map.
std::string PropertyMap::getString(std::string name) const {
	std::transform(name.begin(), name.end(), name.begin(), (int(*)(int))std::tolower); // Horrible!
	std::map<std::string, std::string>::const_iterator it = map.find(name);
	if ( it== map.end()) {
		bad_input("PropertyMap: lookup of non-existant key ", name);
	}
	return it->second;
}

/// Convert a string to a double, allowing for the possibility of INF
double convertToDouble(const std::string& s) {
	char* endPtr = NULL;
	double result = 0;
	if (s == "INF") {
		result = std::numeric_limits<double>::infinity();
	}
	else {
		result = strtod(s.c_str(), &endPtr);
		if (*endPtr) {
			bad_input("PropertyMap: error converting a string to a double", s);
		}
	}
	return result;
}

/// Return the value of the named entry in the property map, 
/// interpreted as a double value, e.g. x = 123.457.
/// The string value INF returns a double-precision infinity.
double PropertyMap::getDouble(std::string name) const {
	return convertToDouble(getString(name));
}

/// Return the value of the named entry in the property map, 
/// interpreted as a size_t value, e.g. x = 123457
size_t PropertyMap::getSizeT(std::string name) const {
	std::string s = getString(name);
	char* endPtr = NULL;
	size_t result = (size_t) strtoul(s.c_str(), &endPtr, 10);
	if (*endPtr) {
		bad_input("PropertyMap: error converting value of property ", name);
	}
	return result;
}

/// Return the value of the named entry in the property map, 
/// interpreted as a signed int value, e.g. x = -123457
int PropertyMap::getInt(std::string name)  const {
	return (int) getLong(name);
}

/// Return the value of the named entry in the property map, 
/// interpreted as a signed long, e.g. x = -1234577654321
long PropertyMap::getLong(std::string name)  const {
	std::string s = getString(name);
	char* endPtr = NULL;
	int result = strtol(s.c_str(), &endPtr, 10);
	if (*endPtr) {
		bad_input("PropertyMap: error converting value of property ", name);
	}
	return result;
}

/// Return the value of the named entry in the property map, 
/// interpreted as a boolean value, e.g. x = false or x = true
bool PropertyMap::getBool(std::string name)  const {
	std::string s = getString(name);
	if (s != "true" && s != "false") {
		bad_input("PropertyMap: error converting value of property ", name);
	}
	return s == "true";
}

/// Strip braces at start and end of string, altering parameter.
/// Return false if braces not present (ignoring spaces).
bool stripBraces(std::string& s) {
	s = stripCommentsAndSpaces(s);
	if (s.size() == 0 || s[0] != '{' || s[s.size()-1] != '}') {
		return false;
	}
	else {
		s = s.substr(1, s.size()-2);
		return true;
	}
}

/// Extract from the parameter string the first "{...}" sequence plus a possible terminating comma.
/// Parameter string is modified, return value is the extracted bit.
/// Returns the empty string if no such leading string can be found.
std::string extract_first_row(std::string& s) {
	s = stripCommentsAndSpaces(s);
	std::string extracted_bit;
	if (s.size() != 0 && s[0] == '{') {
		size_t pos = s.find('}');
		if (pos != std::string::npos) {
			extracted_bit = s.substr(0, pos+1);
			s = stripCommentsAndSpaces(s.substr(pos+1));
			if (s.size() != 0 && s[0] == ',') {
				s = s.substr(1);
			}
		}
	}
	return extracted_bit;
}

/// Interpret the given string parameter as a vector of doubles, defined using
/// the usual C syntax for array initialisers but with the addition of the special
/// value INF to denote infinity.
std::vector<double> parse_array(std::string& s) {
	if (!stripBraces(s)) {
		bad_input("PropertyMap: Missing braces for 1D array ", s);
	}
	std::vector<double> result;
	std::string token;
	while ((token = nextDoubleToken(s)) != "") {
		result.push_back(convertToDouble(token));
		s = trim(s);
		if (s != "") {
			if (s[0] != ',') {
				bad_input("PropertyMap: Wrong syntax for array of doubles", s);
			}
			s = trim(s.substr(1));
		}
	}
	return result;
}

/// Return the value of the named entry in the property map,
/// interpreted as a 2D array (type vector< vector<double> >
std::vector<double> PropertyMap::getDoubleArray(std::string name) const {
	std::string s = getString(name);
	return parse_array(s);
}


/// Return the value of the named entry in the property map,
/// interpreted as a 2D array (type vector< vector<double> >
/// using the usual C initialisation syntax.
std::vector< std::vector<double> > PropertyMap::getDoubleArray2D(std::string name) const {
	std::string s = getString(name);
	if (!stripBraces(s)) {
		bad_input("PropertyMap: error converting value of 2D array property ", name);
	}

	std::vector< std::vector<double> > result;
	std::string row;
	while ((row = extract_first_row(s)) != "") {
		std::vector < double > array_row = parse_array(row);
		result.push_back(array_row);
	}
	return result;

}
