/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

Authors:    Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
			Telecomunication Networks Group
			Politecnico di Torino
			Torino, Italy
			mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file
    Define the PropertyMap class. A property map is a map from string
	to string but with access methods that allow a named property to be
	reads as an integer, a double or a string. It is initialised from
	a text file containing lines like:

	MAX_BANDWIDTH = 20.75   // This is a comment.

	A global object of this class, called 'config' is used to
	define simulation parameters. The file 'config.txt' must be in the
	current directory when the program is run. If not, the simulator aborts
	with a fatal error.

	Property names are now case-insensitive.

    The file 'config.example.txt' is part of the distribution. It can
    be renamed to 'config.txt' to have a working copy.
 */

#ifndef _PROPERTY_MAP_H_
#define _PROPERTY_MAP_H_

#include <string>
#include <map>
#include <fstream>
#include <vector>
#include <set>

/// The PropertyMap class
class PropertyMap {
	static std::set<std::string> known_keys;
	std::map<std::string, std::string> map;

public:
	void load(const char* config_filename);	///< Load map from file. See file 'config.txt' for a sample input
	void dump(FILE* file);			///< Dump the contents of the map to given file
	void process_command_line_args(size_t argc, char* argv[]);  ///< Process the command line args as property values
	bool has_key(std::string name) const;		///< True iff given symbol exists in the map
	double getDouble(std::string name) const; ///< Return the value of the property 'name', interpreted as a double
	size_t getSizeT(std::string name) const;  ///< Return the value of the property 'name', interpreted as an unsigned int 
	int getInt(std::string name) const;       ///< Return the value of the property 'name', interpreted as a signed int 
	long getLong(std::string name) const;     ///< Return the value of the property 'name', interpreted as a long
	bool getBool(std::string name) const;	  ///< Return the value of the property 'name', interpreted as a bool
	std::vector<double> getDoubleArray(std::string name) const; ///< Return the value of the property 'name', interpreted as a 1D array of doubles
	std::vector< std::vector<double> > getDoubleArray2D(std::string name) const; ///< Return the value of the property 'name', interpreted as a 1D array of doubles
	std::string getString(std::string name) const;  ///< Return the raw property 'name.

private:
	void process_line(std::string line, bool allow_overwrite = false);  ///< Process a line from the file or a command line arg
};

#endif
