# A python program (using numpy and matplotlib) that provides a range of
# plotting and analysis capabilities on output data from the simulator.
# [Numpy and matplotlib are powerful
# matlab-like tools, capable of doing fast efficient numerical
# calculations and producing excellent graphical output (2D only).]
# Note that the graphs in the output window can be saved to files
# in a wide variety of formats using the toolbar below the graph.

# RJL, October 2008, December 2008


usingIDLE = 1     # Not sure if this is necessary. Seems OK to leave it at 1??

#--- modules ---

import matplotlib
matplotlib.use('TkAgg')
import matplotlib.pyplot as plt
import tkMessageBox, tkSimpleDialog, tkFileDialog
import Tkinter as tk
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2TkAgg
from matplotlib.figure import Figure
from numpy import arange, sin, pi
import numpy as np
import tkFont
import os


# The histogram dialog class is used to bring up a dialog box
# through which the user can provide parameters for the histogram plots.
# It's recommended that you use the default values first then, after
# looking at the output, decide which values need to be changed from
# default.
# To understand this class, see the excellent Tkinter document
# an-introduction-to-tkinter.pdf (available in many places on the web)
class HistogramDialog(tkSimpleDialog.Dialog):

    # Constructor
    def __init__(self,
                 parent,
                 fields):
        self.fields = fields  # Dictionary of fields and values with which to build dialogue
        self.entries = {}
        self.cancelled = True
        tkSimpleDialog.Dialog.__init__(self, parent)

    # Body, called when setting up the dialog
    def body(self, master):
        
        def makeEntry(master, row, column, string):
            entry = tk.Entry(master)
            entry.grid(row=row, column=column)
            entry.insert(0, string)
            return entry

        # To order the fields, I have a set of possible keys which, if present, will be in the
        # same order as in that list. All other keys will follow, in alphabetical order.
        
        knownKeys = ["title","xLabel","yLabel","numBins", "xmin","xmax","ymin","ymax", "peer0", "peer1"]
        requiredKeys = self.fields.keys()
        orderedKeys = [k for k in knownKeys if k in requiredKeys]
        otherKeys = list(set(requiredKeys) - set(orderedKeys))
        otherKeys.sort()
        
        tk.Label(master, text="Leave fields blank to use defaults").grid(row=0, column=0, columnspan=2)
        row = 1
        for key in orderedKeys + otherKeys:
            tk.Label(master, text=key).grid(row=row)
            self.entries[key] = makeEntry(master, row, 1, str(self.fields[key]))
            row += 1

    # Called when the box is closed, to set the fields ready for use
    # by the user.
    def apply(self):
        for key in self.fields:
            self.fields[key] = self.entries[key].get()
        self.cancelled = False

    def isok(self):
        return not self.cancelled

    def getFields(self):
        return self.fields

class LegendLocDialog(tkSimpleDialog.Dialog):

    # Constructor
    def __init__(self,
                 parent):
        self.xEntry = self.yEntry = None
        tkSimpleDialog.Dialog.__init__(self, parent)

    # Body, called when setting up the dialog
    def body(self, master):
        tk.Label(master, text="Leave fields blank to use default legend position").grid(row=0, column=0, columnspan=2)
        tk.Label(master, text="x: ").grid(row=1, column=0)
        self.xEntry = tk.Entry(master)
        self.xEntry.grid(row=1, column = 1)
        tk.Label(master, text="y: ").grid(row=2, column=0)
        self.yEntry = tk.Entry(master)
        self.yEntry.grid(row=2, column=1)

    def apply(self):
        xs = self.xEntry.get()
        ys = self.yEntry.get()
        try:
            self.result = (float(xs), float(ys))
        except:
            self.result = None

class DegreePlotDialog(tkSimpleDialog.Dialog):
    def __init__(self, parent, title, maxChunk, inclSource):
        self.xEntry = self.yEntry = None
        self.myTitle = title
        self.maxChunk = maxChunk
        self.inclSource = inclSource
        tkSimpleDialog.Dialog.__init__(self, parent)

    def body(self, master):
        tk.Label(master, text="Title: ").grid(row=0,column=0)
        self.titleEntry = tk.Entry(master)
        self.titleEntry.insert(0, self.myTitle)
        self.titleEntry.grid(row=0, column=1)
        tk.Label(master, text="maxChunk: ").grid(row=1, column=0)
        self.maxChunkEntry = tk.Entry(master)
        self.maxChunkEntry.insert(0, str(self.maxChunk))
        self.maxChunkEntry.grid(row=1, column=1)
        tk.Label(master, text="Legend x: ").grid(row=2, column=0)
        self.xEntry = tk.Entry(master)
        self.xEntry.grid(row=2, column = 1)
        tk.Label(master, text="Legend y: ").grid(row=3, column=0)
        self.yEntry = tk.Entry(master)
        self.yEntry.grid(row=3, column=1)
        self.inclSourceVar = tk.IntVar()
        tk.Checkbutton(master,text="Include source", variable=self.inclSourceVar).grid(row=4,column=0)

    def apply(self):
        try:
            try:
                coords = (float(self.xEntry.get()), float(self.yEntry.get()))
            except:
                coords = None
            self.result = (self.titleEntry.get(), int(self.maxChunkEntry.get()),\
                           self.inclSourceVar.get() == 1, coords)
        except:
            self.result = None
        
# uiClass is essentially the main class for the program.
# It currently contains all the code to set up the GUI and also
# all the code that executes the GUI commands.

class uiClass:
    def __init__(self, master):

        # create menubar

        self.parent = master
        menubar = tk.Menu(master) 
        master.config(menu=menubar)

        # create dropdown menus
    
        graphsMenu = tk.Menu(menubar, tearoff=0)
        fileMenu = tk.Menu(menubar, tearoff=0)
        chunkMenu = tk.Menu(menubar, tearoff=0)
        linksMenu = tk.Menu(menubar, tearoff=0)
        
        fileMenu.add_command(label='Folder...', underline=0, command=self.selectFolder)
        fileMenu.add_command(label='Exit', underline=1, command=self.quit)
        graphsMenu.add_command(label='Chunk loss freq', underline=0, command=self.lostChunkStats)
        graphsMenu.add_command(label='Chunk loss by peer',command=self.chunkLossesByPeer)
        graphsMenu.add_command(label='Chunk delays (old)', command=self.chunkDelayStats)
        graphsMenu.add_command(label='Chunk delays (new)', command=self.chunkDelayStatsNew)
        graphsMenu.add_command(label='Out-degree(t)', command=self.outdegree)
        graphsMenu.add_command(label='In-degree(t)', command=self.indegree)
        graphsMenu.add_command(label='Delay(chunk_num)', command=self.delayVersusT)
        chunkMenu.add_command(label='Lost chunks', command=self.lostChunks)
        chunkMenu.add_command(label='Trace chunk', underline=0, command=self.traceChunk)
        chunkMenu.add_command(label='Delay stats', command=self.delayStats)
        linksMenu.add_command(label='Analyse', command=self.analyseLinks)

        # Add dropdowns to menubar
        
        menubar.add_cascade(label='File', menu=fileMenu, underline=0)
        menubar.add_cascade(label='Graphs', menu=graphsMenu, underline=0)
        menubar.add_cascade(label='Chunks', menu=chunkMenu, underline=0)
        menubar.add_cascade(label='Links', menu=linksMenu)

        # Create the main GUI window
        
        self.frame = tk.Toplevel(relief='ridge', borderwidth=2, menu=menubar)
        self.frame.resizable(1, 1)
        self.frame.protocol('WM_DELETE_WINDOW', self.quit)

        # Create the matplotlib drawing canvas and navigation toolbar

        leftPane = tk.Frame(self.frame)
        leftPane.pack(side=tk.LEFT, fill=tk.BOTH, expand=1)
        self.figure = Figure(figsize=(8,6), dpi=100)
        self.canvas = FigureCanvasTkAgg(self.figure, master=leftPane)
        self.canvas.show()
        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar2TkAgg( self.canvas, leftPane)
        self.toolbar.update()
        self.canvas._tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=1)

        # Initialise state variables

        self.folder = "./output"
        self.widgets = {}  # Dictionary of widgets and their cursors
        self.chunkDelayFields = dict(  # Dictionary for chunk delay histogram dialogue
            title = "Chunk delay histogram",
            xLabel = "Delay (secs)",
            yLabel = "Frequency",
            numBins = 100,
            xmin = "",
            xmax = "",
            ymin = "",
            ymax = "",
            peer0 = 0,
            peer1 = 0)
        
        self.newChunkDelayFields = dict(  # Dictionary for new chunk delay histogram dialogue
            title = "Chunk delay histogram",
            xLabel = "Delay (secs)",
            yLabel = "Frequency",
            numToMerge = 1,
            xmin = "",
            xmax = "",
            ymin = "",
            ymax = "",
            class0 = 0,
            class1 = 4)

        self.chunkLossHistoFields = dict(  # Dictionary for chunk-loss histogram dialogue
            title = "Chunk lost statistics",
            xLabel = "Num chunks lost",
            yLabel = "Num peers with this loss",
            numBins = 100,
            xmin = "",
            xmax = "",
            ymin = "",
            ymax = "")
        
        self.peerChunkLossFields = dict(  # Dictionary for chunk-loss bar graph dialogue
            xmin=0, xmax="",
            ymin=0, ymax="",
            title="Chunk loss by peer",
            xLabel="Peer ID",
            yLabel="Chunks lost")
        self.text = None
        self.fileCache = {}  # (timeStamp, fileData) cache maintained by readNumbers



    # Add a message output area with its own scrollbar
    def addTextPane(self):
        rightPane = tk.Frame(self.frame)
        rightPane.pack(side=tk.RIGHT, fill=tk.Y)
        scrollbar = tk.Scrollbar(rightPane)
        scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        textFont = tkFont.Font(family="Courier", size=9)
        self.text = tk.Text(rightPane, yscrollcommand=scrollbar.set, width=60, font=textFont )
        self.text.pack(side=tk.LEFT, fill=tk.BOTH)
        scrollbar.config(command=self.text.yview)


    # FILE MENU COMMANDS
    # ==================
    # Command-handler for selecting the folder containing the stats files
    # from a simulator run.
    
    def selectFolder(self):
        folder = tkFileDialog.askdirectory()
        if folder != None:
            self.folder = folder
        self.fileCache.clear();
        self.frame.focus_set()

    # 'Quit' command handler
    
    def quit(self):
        if usingIDLE:
            self.frame.destroy()
            root.quit()
        else:
            self.frame.quit()
            root.quit()

    # PLOT (aka STATS) MENU COMMANDS
    # ==============================
    # Command-handler for displaying a histogram of lost-chunk stats
    # i.e., the number of peers that lost no chunks, the number losing 1, 2, ...
    def lostChunkStats(self):
        self.busy()
        filedata = self.readNumbers("Total_chunk_losses_per_peer.out")
        if filedata != None:
            paramDlg = HistogramDialog(self.frame, self.chunkLossHistoFields)
            if paramDlg.isok():
                self.chunkLossHistoFields = paramDlg.getFields()
                self.plotHistogram(filedata, self.chunkLossHistoFields)      
        self.notBusy()

    # Command-handler for displaying a histogram of chunk-delay statistics
    # This is the old version, which uses the expensive Chunk_delay.out file
    def chunkDelayStats(self):
        self.busy()
        delays = self.readNumbers("Chunk_delay.out");
        if delays != None:
            if self.chunkDelayFields['peer1'] == 0:
                self.chunkDelayFields['peer1'] = len(delays[0]) - 2
            paramDlg = HistogramDialog(self.frame, self.chunkDelayFields)
            if paramDlg.isok():
                self.chunkDelayFields = paramDlg.getFields()
                p0 = self.intField(self.chunkDelayFields,'peer0')
                p1 = self.intField(self.chunkDelayFields,'peer1')
                allDelays = np.array(delays[0:,p0+1:p1+2].flat)
                self.plotHistogram(allDelays, self.chunkDelayFields) 
        self.notBusy()

    # Command-handler for displaying a histogram of chunk-delay statistics
    # This is the new version, using the per-peer chunk delay histogram data
    def chunkDelayStatsNew(self):
        self.busy()
        delays = self.readNumbers("DelayHistograms.out", returnType="list");
        if delays != None:
            numBars = int(delays[0][0])
            numPeers = int(delays[0][1])
            barWidth = delays[0][2]
            paramDlg = HistogramDialog(self.frame, self.newChunkDelayFields)
            if paramDlg.isok():
                fields = paramDlg.getFields()
                self.newChunkDelayFields = fields
                c0 = self.intField(fields,'class0')
                c1 = self.intField(fields,'class1')
                numLost = 0

                allDelays = np.zeros(numBars)
                for peer in range(0, numPeers):
                    (id, peer_class, lost) = delays[1 + peer*2]
                    if peer_class in range(c0, c1+1):
                        allDelays += np.array(delays[2 + peer*2])
                        numLost += lost

                numToMerge = self.intField(fields,'numToMerge')
                numBins = int(numBars / numToMerge)
                mergedBins = np.zeros(numBins)
                for bin in range(0, numBins):
                    for i in range(0, numToMerge):
                        mergedBins[bin] += allDelays[bin * numToMerge + i]
                axes = self.figure.gca()
                axes.clear()
                axes.set_xlabel(fields['xLabel'])
                axes.set_ylabel(fields['yLabel'])
                axes.set_title(fields['title'])
                lefts = [-1] + [i*barWidth*numToMerge for i in range(0, numBins)]
                data = [numLost] + list(mergedBins)
                w = barWidth*numToMerge
                axes.bar(lefts, data, width=w, fc='green', ec='green')
                axes.set_xlim(xmin=self.floatField(fields, 'xmin'))
                axes.set_xlim(xmax=self.floatField(fields, 'xmax'))
                axes.set_ylim(ymin=self.floatField(fields, 'ymin'))
                axes.set_ylim(ymax=self.floatField(fields, 'ymax')) 
        self.notBusy()



    # Command handler for displaying chunks lost by each peer
    def chunkLossesByPeer(self):
        self.busy()
        filedata = self.readNumbers("Total_chunk_losses_per_peer.out")
        if filedata != None:
            data = [col[0] for col in filedata]
            self.figure.clear()
            self.peerChunkLossFields['ymax'] = max(data)
            self.peerChunkLossFields['xmax'] = len(data)
            paramDlg = HistogramDialog(self.frame, self.peerChunkLossFields)
            if paramDlg.isok():
                axes = self.figure.gca()
                fields = paramDlg.getFields()
                self.peerChunkLossFields = fields
                axes.set_xlabel(fields['xLabel'])
                axes.set_ylabel(fields['yLabel'])
                axes.set_title(fields['title'])
                axes.bar(range(0, len(data)), data, width=1, fc='green', ec='green')
                axes.set_xlim(xmin=self.floatField(fields, 'xmin'))
                axes.set_xlim(xmax=self.floatField(fields, 'xmax'))
                axes.set_ylim(ymin=self.floatField(fields, 'ymin'))
                axes.set_ylim(ymax=self.floatField(fields, 'ymax'))
                axes.grid(True, linestyle="-", color="0.7")
        self.notBusy()

    def plotDegree(self, colOffsets, labels, title, symbols):
        self.busy()
        filedata = self.readNumbers("DegreeOverTime.out")
        if filedata != None:
            lastChunk = int(filedata[-1][0])
            params = DegreePlotDialog(self.frame, title, lastChunk, False).result
            if params != None:
                (title, maxChunk, inclSource, legCoords) = params
                newFirstRow = filedata[0].copy()
                newFirstRow[0] = 0
                filedata = np.insert(filedata, 0, newFirstRow, axis=0) # Hack to replicate the row zero data
                self.figure.clear()
                axes = self.figure.gca()
                axes.set_xlabel("Chunk")
                axes.set_ylabel("Degree")
                axes.set_title(title)
                if inclSource:
                    first = 0
                else:
                    first = 1
                xs = [col[0] for col in filedata if col[0] <= maxChunk]
                plots = [[col[colOffset] for col in filedata if col[0] <= maxChunk] for colOffset in colOffsets]
                currSym = first
                for plot in plots[first:]:
                    axes.plot(xs, plot, symbols[currSym], label=labels[currSym])
                    currSym = (currSym + 1) % len(symbols)
                if (legCoords == None):
                    axes.legend(loc='best')
                else:
                    axes.legend(loc=legCoords)
                axes.set_ylim(ymin=0)
                axes.grid(True, linestyle="-", color="0.7")
                self.canvas.show()
        self.notBusy()
        
    # Command handler for outdegree command, which displays the variation of
    # out-degree over time for the various different peer classes
    def outdegree(self):
        symbols = ["k:","b-","m--","k-.", "y-"] #"ro-","b>-","gx-","yh-","mD-","cs-","k+-","rx-"]
        self.plotDegree([4 + 3*i for i in range(0,5)],\
                        ["Source","Out1","Out2","Out3","Out4"], "Out-degree(t)", symbols)

    # Command handler for the indegree command, which displays the variation of
    # in-degree over time for the various different peer classes
    def indegree(self):
        symbols = ["k:","b-","m--","k-.","y-"]
        self.plotDegree([3 + 3*i for i in range(0,5)], \
                        ["Source","In1","In2","In3","In4"], "In-degree(t)", symbols)

    # Command handler for the delayVersusT command, which displays a plot of
    # average chunk delay (over all peers) versus chunk number.
    def delayVersusT(self):
        self.busy()
        delays = self.readNumbers("Chunk_delay.out", returnType="list" );
        if delays != None:
            nChunks = len(delays)
            firstChunkNum = delays[0][0]
            numToAverage = tkSimpleDialog.askinteger("Window size", "Number of chunks to average")
            lastChunk = int(delays[-1][0])
            maxChunk = tkSimpleDialog.askinteger("Up to what chunk?", "Up to what chunk?", initialvalue=lastChunk)

            nPeers = len(delays[0])-2
            (maxClass, lookupTable) = self.getPeerClasses("%s/Upload_rate_peer.out" % self.folder);
            maxChunkBin = int(nChunks/numToAverage);
            #print "nChunks = ", nChunks, ", maxClass = ", maxClass
            classDelayTotals = np.zeros((maxClass+1, maxChunkBin+1))
            classDelayCounts = np.zeros((maxClass+1, maxChunkBin+1))
            for row in delays:
                if row[0] > maxChunk and int(row[0] - firstChunkNum) % numToAverage == 0:
                    break
                for peer in range(1, nPeers+1):
                    peerClass = lookupTable[peer]
                    chunkBin = int((row[0] - firstChunkNum)/numToAverage)
                    delay = row[peer+1]
                    classDelayTotals[peerClass][chunkBin] += delay
                    classDelayCounts[peerClass][chunkBin] += 1
                    classDelayTotals[0][chunkBin] += delay
                    classDelayCounts[0][chunkBin] += 1
            meanDelays = classDelayTotals/classDelayCounts
            self.figure.clear()
            axes = self.figure.gca()
            axes.set_xlabel("Chunk")
            axes.set_ylabel("Delay")
            #axes.set_title("Mean chunk delay versus chunk number")
            xs = [firstChunkNum + chunkBin*numToAverage for chunkBin in range(0, maxChunkBin+1)]
            symbols = ["k:","b-","m--","k-.","y-"]
            labels = ["All", "Class1", "Class2", "Class3", "Class4"]
            currSym = 0
            for cls in range(0, maxClass+1):
                axes.plot(xs, meanDelays[cls], symbols[currSym], label=labels[currSym])
                currSym = (currSym + 1) % len(symbols)
            coords = LegendLocDialog(self.frame).result
            if (coords == None):
                axes.legend(loc='best')
            else:
                axes.legend(loc=coords)
            axes.set_ylim(ymin=0)
            axes.grid(True, linestyle="-", color="0.7")
            self.canvas.show()            
        self.notBusy()
        

    # CHUNK MENU COMMANDS
    # ===================
    def lostChunks(self):
        self.busy()
        arrivalTimes = self.readNumbers("Absolute_time_arrival.out");
        if arrivalTimes != None:
            lostChunks = []
            rowNum =  arrivalTimes[0][0]
            if rowNum != 0:
                print "Warning: no data prior to chunk #", rowNum
            for row in arrivalTimes:
                chunkNum = row[0]
                assert chunkNum == rowNum
                count = 0
                for arrivalTime in row[1:]:
                    if arrivalTime < 0:
                        count += 1
                if count != 0:
                    lostChunks.append((chunkNum, count))
                rowNum += 1
            self.printline(" Chunk  NumLosses")
            for lc in lostChunks:
                self.printline("%5d%10d" % lc)
            self.printline((15 * '-') + "\n")
        self.notBusy()
        

    def traceChunk(self):
        chunkNum = tkSimpleDialog.askinteger("Chunk num", "Enter number of chunk to trace")
        if chunkNum != None:
            self.busy()
            arrivalTimes = self.readNumbers("Absolute_time_arrival.out");
            if arrivalTimes != None:
                assert arrivalTimes[chunkNum][0] == chunkNum
                times = arrivalTimes[chunkNum][1:]
                chunkArrivals = []
                for i in range(0, times.shape[0]):
                    if times[i] >= 0:
                        chunkArrivals.append((i, times[i]))
                chunkArrivals.sort(lambda x,y: cmp(x[1],y[1]))
                self.printline("Trace for chunk %d" % chunkNum)
                self.printline("Node    ArrTime")
                for arrival in chunkArrivals:
                    self.printline("%5d%10.4f" % arrival)
                self.printline("\n")
            self.notBusy()

    # Min max median & 95%ile estimates of delay stats from histogram data,
    # overall and by class
    def delayStats(self):
        delays = self.readNumbers("DelayHistograms.out", returnType="list");
        if delays != None:
            numBars = int(delays[0][0])
            numPeers = int(delays[0][1])
            barWidth = delays[0][2]
            allDelays = np.zeros(numBars)
            totLosses = 0
            lossesByClass = {}
            delaysByClass = {}
            for peer in range(0, numPeers):
                (id, peer_class, lost) = delays[1 + peer*2]
                binCounts = np.array(delays[2 + peer*2])
                allDelays += binCounts
                totLosses += lost
                if peer_class in delaysByClass:
                    delaysByClass[peer_class] += binCounts
                    lossesByClass[peer_class] += lost
                else:
                    delaysByClass[peer_class] = binCounts
                    lossesByClass[peer_class] = lost
            self.printline("Class  Lost  Loss%  Min   Max  Mean Median 95%ile")
            self.printStats(barWidth, "All", totLosses, allDelays)
            for c in delaysByClass:
                self.printStats(barWidth, "%d" % c, lossesByClass[c], delaysByClass[c])
            self.printline("")

    # Stats-printing support method for delayStats. Computes actual statistics given
    # an array of histogram bin counts (assumed to start at zero) and a bin width.
    def printStats(self, binWidth, className, lost, bins):
        binCentre = binWidth/2
        numNotLost = np.sum(bins)
        halfNum = numNotLost/2
        percent95 = numNotLost*0.95
        percentile95 = None
        min = None
        median = None
        max = 0
        totl = 0
        cnt = 0
        
        for bin in bins:
            if bin > 0:
                if min == None:
                    min = binCentre
                max = binCentre
                cnt += bin
                totl += bin * binCentre
                if median == None and cnt >= halfNum:
                    median = binCentre
                if percentile95 == None and cnt >= percent95:
                    percentile95 = binCentre
            binCentre += binWidth

        cnt += lost
        self.printline("%5s %5d %5.3f%% %5.3f %5.3f %5.3f %5.3f %5.3f" % \
            (className, lost, 100.0*lost/cnt, min, max, totl/numNotLost, median, percentile95))

        

    # LINKS MENU COMMAND
    # ==================
    def printDegrees(self, lis, maxClass):
        self.printline("  Min  Max   Avg     SD   Total   Num")
        for c in range(0,maxClass+1):
            data = np.array(lis[c])
            self.printline("%5d %5d %6.2f %6.2f %6d %5d" % \
                (data.min(), data.max(), data.mean(), data.std(), data.sum(), len(lis[c])))

    # Build a lookup-table of peer-classes by reading the given upload-rate file.
    # Returns (maxClassNum, lookupTable) where lookupTable[i] is the class of peer #i
    def getPeerClasses(self, filename):
        maxClass = -1
        try:
            peers_file = open(filename, "r")
        except IOError:
            tkMessageBox.showerror(title="No file", message = "File %s not found" % filename)
            return
        lines = peers_file.readlines()
        peers_file.close()
        peer_class=[]
        for line in lines:
            bits = line.split()
            assert int(bits[0]) == len(peer_class)
            itsClass = int(bits[2])
            if itsClass > maxClass:
                maxClass = itsClass
            peer_class.append(itsClass)
        return (maxClass, peer_class)

    # Return a array giving the number of times each peer appears in the peer_list
    def get_frequencies(self, peer_list, numPeers):
        counts = [0 for peer in range(0, numPeers)]
        for p in peer_list:
            counts[p] += 1
        return counts

    # Split the list freqs (frequencies of occurrence of each peer in some list),
    # into a list of sublists, one for each class of peer according to the given
    # classes array.
    def classify(self, freqs, classes, maxClass):
        degrees = [[] for c in range(0, maxClass+1)]
        for i in range(0, len(classes)):
            degrees[classes[i]].append(freqs[i])
        return degrees
        
        
    def analyseLinks(self):
        (maxClass, peer_class) = self.getPeerClasses("%s/Upload_rate_peer.out" % self.folder)
        num_peers = len(peer_class)
        filename = "%s/LinksAtEnd.out" % self.folder
        try:
            links_file = open(filename, "r")
        except IOError:
            tkMessageBox.showerror(title="No file", message = "File %s not found" % filename)
            return
        lines = links_file.readlines()
        links_file.close()

        counts = {}
        totalLinks = 0
        for i in range(0,maxClass+1):
            for j in range(0,maxClass+1):
                counts[(i,j)] = 0

        last_src = -1
        num_out = 0
        source_occurrences = []
        dest_occurrences = []

        for line in lines[1:]:
            bits = line.split()
            src = int(bits[1])
            dest = int(bits[3])
            source_occurrences.append(src)
            dest_occurrences.append(dest)
            counts[(peer_class[src], peer_class[dest])] += 1
            totalLinks += 1

        out_degrees = self.get_frequencies(source_occurrences, num_peers)
        in_degrees = self.get_frequencies(dest_occurrences, num_peers)

        out_degrees_by_class = self.classify(out_degrees, peer_class, maxClass)
        in_degrees_by_class = self.classify(in_degrees, peer_class, maxClass)

        self.printline("Total links: %d, Avge degree: %.1f" % (totalLinks, totalLinks/float(num_peers)))

        self.printline("Out-degrees:")
        self.printDegrees(out_degrees_by_class, maxClass)

        self.printline("")

        self.printline("In-degrees:")
        self.printDegrees(in_degrees_by_class, maxClass)

        self.printline("")
        
        self.printline("Links from class row to class column")
        for csrc in range(0,maxClass+1):
            countString = "%s: " % csrc
            for cdest in range(0,maxClass+1):            
                countString += "%6d" % counts[(csrc,cdest)]
            self.printline(countString + " Sum:%6d" % np.array(out_degrees_by_class[csrc]).sum() )
        self.printline("")

        self.printline("Link percentages from class row to class column")
        for csrc in range(0,maxClass+1):
            numOutLinks = np.array(out_degrees_by_class[csrc]).sum()
            countString = "%s: " % csrc
            if numOutLinks == 0:
                countString += "No links"
            else:
                for cdest in range(0,maxClass+1):
                    percent = counts[(csrc,cdest)] * 100 / float(numOutLinks)
                    countString += "%5.1f " % percent
            self.printline(countString)
        self.printline("")


    # INTERNAL SUPPORT METHODS
    # ========================
    # Service method -- reads a file of numbers into a 2D numpy array of floats.
    # Params: filename, relative to current folder.
    # Returns: 2D numpy array of floats.
    # Caches previously read files so a given file should only be read once
    # (unless folder command is used to change out of a folder and back in --
    # cache is then cleared in the interests of memory consumption).
    def readNumbers(self, filename, returnType="numpy"):
        fname = self.folder + "/" + filename;
        try:
            f = open(fname, "r")
        except IOError:
            f = None
        if f == None:
            tkMessageBox.showerror(title="No file", message = "File %s not found" % fname)
            return None

        last_changed_at = os.stat(fname).st_mtime
        if not (fname in self.fileCache) or self.fileCache[fname][0] < last_changed_at:
            lines = f.readlines()
            data = []
            for line in lines:
                thisLine = []
                bits = line.split()
                for bit in bits:
                    thisLine.append(float(bit))
                data.append(thisLine)
            if returnType == "numpy":
                values = np.array(data)
            else:
                values = data
            self.fileCache[fname] = (last_changed_at, values)  # (timeStamp, data)
        f.close()
        return self.fileCache[fname][1]

    # Two utility methods for getting values out of a histogram-fields dictionary.
    def intField(self, fields, key):
        s = fields[key]
        if s == "":
            return None
        else:
            return int(s)
        
    def floatField(self, fields, key):
        result = None
        s = fields[key].strip()
        if s != "":
            result = float(s)
        return result
            
    # Plot a histogram
    def plotHistogram(self, data, fields):
        self.figure.clear()
        axes = self.figure.gca()
        axes.set_xlabel(fields['xLabel'])
        axes.set_ylabel(fields['yLabel'])
        axes.set_title(fields['title'])
        numBins = self.intField(fields, 'numBins')
        xmin = self.floatField(fields, 'xmin')
        xmax = self.floatField(fields, 'xmax')
        ymin = self.floatField(fields, 'ymin')
        ymax = self.floatField(fields, 'ymax')

        if xmin != None and xmax != None:
            binWidth = (xmax - xmin) / float(numBins)
            binSet = [ xmin + i*binWidth for i in range(0, numBins)]
            axes.hist(data, bins=binSet, facecolor='green')
        else:
            axes.hist(data, numBins, facecolor='green')

        if xmin != None:
            axes.set_xlim(xmin=xmin)
        if ymin != None:
            axes.set_ylim(ymin=ymin)
        if xmax != None:
            axes.set_xlim(xmax=xmax)
        if ymax != None:
            axes.set_ylim(ymax=ymax)
        axes.grid(True, linestyle="-", color="0.7")
        self.canvas.show()
        
    
    # Print a given line of text at the end of the text output window
    def printline(self, s):
        if self.text == None:
            self.addTextPane()
        self.text.insert(tk.END, s + "\n")
        self.text.yview_moveto(1.0)
        
    # Attach watch cursor to all widgets.
    def busy(self, widget=None):
        if widget is None:
            w = self.frame
        else:
            w = widget

        if not self.widgets.has_key(str(w)):
            try:
                # attach cursor to this widget
                cursor = w.cget("cursor")
                if cursor != "watch":
                    self.widgets[str(w)] = (w, cursor)
                    w.config(cursor="watch")
            except TclError:
                pass

        for w in w.children.values():
            self.busy(w)

    def notBusy(self):
        # restore cursors
        for (w, cursor) in self.widgets.values():
            try:
                w.config(cursor=cursor)
            except TclError:
                pass
        self.widgets = {}

# Main body: just set up the main uiClass instance and enter the main loop.
root = tk.Tk()
root.title('SimStats')
root.withdraw() 
mainUI = uiClass(root)
root.mainloop()

