/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file stats.h
	A file containing the declarations of the statistics-gathering classes.
	However, note that the corresponding cpp source file also includes the
	definitions of the Overlay class's statistics methods, such as
	<i>print_statistics</i>, so as to keep all stats gathering, manipulation
	and printing methods in one place.
 */
#ifndef _STATS_H_
#define _STATS_H_
#include <vector>
#include <map>
#include "types.h"

/// The PeerStats class, an instance of which is present in every peer for
/// recording information of a purely statistical nature. Fields of a
/// PeerStats object should <i>not</i> be used to control the diffusion
/// of chunks through the overlay.
class PeerStats {
	static size_t s_first_chunk_to_record;	///< ID number of first chunk to bother recording
	static size_t s_last_chunk_to_record;	///< ID number of last chunk to bother recording
	static bool s_recording_chunk_arrivals; ///< True if every chunk arrival at every host to be recorded
	static size_t s_num_delay_counters;		///< Number of chunk_delay_counters entries
	static double s_delay_histogram_bar_width; ///< Range of delays recorded by each chunk delay cntr
	size_t s_num_chunks_sent;				///< Total count of chunks sent
	size_t s_num_chunks_received;			///< Total count of chunks received
	size_t s_last_chunk_received;			///< Sequence number of last received chunk
	size_t s_num_out_of_seq;				///< Count of times when received chunk seq numbers decrease
	std::vector<size_t> s_delay_counters;	///< Counters for chunk-delay histogram

	// The next field is very expensive in memory if used.
	// To disable it, ensure that ChunkArrivalsFile and ChunkDelaysFile are both
	// commented out in config.txt.
	std::map<ChunkID, Time> s_chunk_arrivals;	///< A map of the arrival times of chunks

	std::vector<PPeer> s_link_deletions;	///< A vector of links deleted
public:
	static void set_first_chunk_to_record(ChunkID first);  ///< Set ID of first chunk of interest
	static void set_last_chunk_to_record(ChunkID first);  ///< Set ID of last chunk of interest
	static void set_num_delay_counters(size_t n);	///< Set how many chunk-delay counters are needed
	static void set_chunk_arrival_recording(bool state);
	static size_t num_delay_counters();		///< The number of delay counters
	static double delay_histogram_bar_width(); ///< Width of delay histogram "bar"

	PeerStats();						///< Constructor: clears all fields
	void clear();						///< Called for all peers by Overlay.clear()
	void record_arrival(PChunk pchunk); ///< Record the arrival of a chunk at the peer
										/// to which this object belongs.
	void record_send(PChunk pchunk);
	void record_link_deletion(PPeer to_whom);    ///< Add an entry to the s_link_deletions field
	Time arrival_time(ChunkID chunk_id);         ///< The arrival time of a given chunk
	size_t num_chunks_received();		     ///< The number of chunks received by the peer
	size_t num_chunks_out_of_seq();
	size_t num_chunks_sent();
	size_t num_chunks_lost();		     ///< Num chunks lost (after first_chunk_to_record)
	double chunk_loss_ratio();                   ///< Per peer chunk loss ratio
	const std::vector<size_t>* delay_counters(); ///< Return a pointer to the delay counters
};

/// The LinksStats class, an instance of which is present in every link for
/// recording information of a purely statistical nature.
class LinksStats {
	size_t s_throughput;	///< Number of chunks sent on this link
public:
	LinksStats();			///< Constructor clears all fields
	void clear();			///< Clear all recorded data (called for all links by Overlay.clear())
	void inc_throughput();
	size_t throughput();
};


#endif
