/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

Authors:    Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
			Telecomunication Networks Group
			Politecnico di Torino
			Torino, Italy
			mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file
    Defines the VideoStream class, which is essentially a list of CHUNKs (for
	which, see the Chunk class). It delivers chunks at the specified
	video rate to the source peer of the overlay.
 */
#ifndef _VIDEO_STREAM_H_
#define _VIDEO_STREAM_H_

#include "types.h"
#include <list>
#include <vector>

enum VideoStreamType { VS_SIMPLE, VS_RINPUT, VS_INPUT, VS_BAD_TYPE };
extern std::string gVideoStreamTypeNames[];



struct VideoStreamCallback {
	void (*callback_func)(size_t chunk_num);	// Function to be called
	size_t chunk_number;		// Chunk number at which function is called
	bool repeats;			// True to call func whenever chunk_id % chunkNumber == 0
	VideoStreamCallback(void (*callback_func)(size_t chunk_num), size_t chunk_number, bool repeats = false):
	callback_func(callback_func), chunk_number(chunk_number), repeats(repeats) {};
};

class VideoStream {
	static std::vector<PChunk> chunk_list;  ///< The vector of all chunks diffused through the overlay
	static size_t current_chunk_id;			///< The ID of the next chunk to be output from the stream
	static size_t num_chunks;				///< The number of chunks in the whole stream
	static size_t num_chunks_w_par;				///< The number of chunks in the whole stream, inlcluding FEC parity chunks
	static std::list<VideoStreamCallback> callbacks;  ///< Callback function list
	static size_t fec_k;
	static size_t fec_n;
	static VideoStreamType vs_type;	///< type of the videostream

	static ChunkSize *SizeArray;
	static ChunkType *gTypeArray;	///< Array for saving chunk type (IDR, I, i, P, p, B or b) red from a pattern file

	static unsigned gFramesForChunk; ///< number of frames (provided fix) for every chunk
	static double gInterFrameTime; ///< inter frame time interval in seconds

	static PChunk chunk_next; ///< parameters of the next chunk


public:
	static void init(VideoStreamType vs_type);

        static PChunk read_next_chunk();

	/// Get the number of chunks in the video stream
	static size_t get_num_chunks(bool w_par = false);

	/// Get the number of the next chunk to send
	static size_t get_next_chunk_num();

	/// Set the number of chunks in the video stream
	static void set_num_chunks(size_t num_chunks);

	/// Set up a callback function to be called just prior to distributing chunk "chunk_num".
	/// repeats is true for repetitive callbacks (whenever chunk_id % chunk_num == 0, except
	/// when chunk_id == 0).
	static void add_callback(void (*func)(size_t chunk_num), size_t chunk_num, bool repeats = false);

	/// Delete all callbacks with the given target function
	static void delete_callback(void (*func)(size_t chunk_num));

	/// Give the next video chunk to the source peer.
	static void next_chunk(POverlay po);

	/// Get a pointer to the chunk with a given ID. This is currently
	/// just a simple indexing operation into the chunk vector.
	/// @return a pointer to the Chunk with the given ID number.
	static PChunk get_chunk(
			size_t chunk_id		///< The ID number of the required chunk
		);

	/// Clear the video stream, freeing all chunks in it.
	static void clear();

	inline static void set_fec_n(size_t n){fec_n=n;}
	inline static size_t get_fec_n(){return fec_n;}
	inline static void set_fec_k(size_t k){fec_k=k;}
	inline static size_t get_fec_k(){return fec_k;}

};

#endif // _VIDEO_STREAM_H_
