/*********************************************************************/
/* Copyright (C) 2008  Politecnico di Torino

 Authors: Marco Mellia, Gloria Bogliolo, Ana Paula Couto, Richard Lobb
  	      Telecomunication Networks Group
  	      Politecnico di Torino
  	      Torino, Italy
  	      mellia@mail.tlc.polito.it, anapaula@tlc.polito.it

 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program.  If not, see <http://www.gnu.org/licenses/>.*/
/*********************************************************************/

/*! @file 
	Define the Window class, a sliding window of chunks
	available for transmission to other peers, that lie within
	a given "playout-delay" of the current simulator time.
	The implementation uses a map from chunk_ID to PChunk for
	the main window and a set of PCHUNKs for recording in-transit
	chunks. The window also maintains a set of chunks marked by the peer
	as being useless. The use of this is entirely over to the individual
	peer: it is has no effect on the rest of the window.
 */


#ifndef _WINDOW_H_
#define _WINDOW_H_

#include <map>
#include <set>
#include "types.h"
#include "chunkinstance.h"
#include "globals.h"

typedef std::map<ChunkID, PChunk> CHUNK_MAP;

/// A ChunkIterator supports iteration through the chunks in a CHUNK_MAP, in 
/// latest-first order. Note that this is the opposite order of the iteration
/// supported by the map's forward iterators (since the map is stored in increasing
/// order of ID, which is the opposite of what we want).
/// This class is really just a wrapper for the map's reverse_iterator class,
/// needed because the latter returns (key, value) pairs whereas we just want
/// to iterate over the values.
class ChunkIterator {
	CHUNK_MAP::reverse_iterator map_iterator;
public:
	inline ChunkIterator(const CHUNK_MAP::reverse_iterator& it) : map_iterator(it) {};
	inline PChunk& operator*() { return map_iterator->second; }
	inline ChunkIterator& operator++() {
		++map_iterator;
		return *this;
	}
	inline ChunkIterator& operator--() {
		--map_iterator;
		return *this;
	}
	inline bool operator!=(const ChunkIterator& other) const {
		return map_iterator != other.map_iterator;
	}
};

/// The window class itself.
class Window
{
	Time playout_delay;				///< Chunks older than this get discarded
	CHUNK_MAP chunks;				///< Chunks we have received that are within the playout delay
        CHUNK_MAP pending;                              ///< Chunks that are on their way
        CHUNK_MAP useless;                              ///< Chunks that the peer has declared as useless
public:
	void insert(PChunk chunk, PPeerStats stats);		///< Insert a chunk into the window

	inline void set_pending(const PChunk chunk) { ///< Record that the specified chunk is on its way
                pending[chunk->id()] = chunk;
	}

	inline void clear_pending(const PChunk chunk) { ///< Clear the "on-it's-way" status of given chunk
	        pending.erase(chunk->id());
	}

	// True iff specified chunk is in the window
	inline bool has_chunk(const PChunk chunk) {
		return chunks.find(chunk->id()) != chunks.end();
	}

	// True iff specified chunk has been marked as useless
	// How this is used is over to the peer. WARNING! Don't assume
	// that just because a peer can't find a neighbour to use a chunk that the chunk
	// is always useless -- a needy neighbour may have been busy!
	inline bool is_useless(const PChunk chunk) {
		return useless.find(chunk->id()) != useless.end();
	}

	/// True iff specified chunk is not in the window, isn't on its way and isn't before the playout delay
	inline bool needs_chunk(PChunk chunk) {
		return !has_chunk(chunk) && pending.find(chunk->id()) == pending.end() &&
				chunk->initiation_time() > gCurrentTime - gPlayoutDelay;
	}

	/// True IFF window is empty
	inline bool empty() {      
		return chunks.empty();
	}

	/// True iff window has chunks pending
	inline bool has_pending_chunks() {
		return !pending.empty();
	}

	/// Mark a chunk as useless
	inline void set_useless(const PChunk chunk) {
		useless[chunk->id()] = chunk;
	}

	void clear();					///< Clear the window

	/// Iterator to the first chunk in the window
	inline ChunkIterator begin() { 
		return ChunkIterator(chunks.rbegin());
	}

	/// (last+1)th PChunk
	inline ChunkIterator end() {
		return ChunkIterator(chunks.rend());
	}	

	/// Latest chunk
	const PChunk latest();

	/// Latest chunk, including pending ones
	const PChunk latest_w_pending();

	void fec_reconstruct(PChunk chunk, PPeerStats stats);
        
        inline int size(){
                return chunks.size();
        }
};


#endif // _WINDOW_H_
